// Copyright 2007 Google, Inc.
// All Rights Reserved.
//
// Author: dbloomberg@google.com (Dan Bloomberg)
// comparison.cc
//
// This program is intended to give examples of two things:
//
//   (1) the types of comparisons you can do
//   (2) the types of output you can derive and display
//
// The output methods shown here include:
//   - Writing a set of images to file using an automatic naming
//     convention and interface of maximal simplicity, so that you
//     can display the results in a program like gthumb.
//   - Accumulating results in an array of images in memory (a Pixa)
//     and then tiling them into a single image that is saved to file.
//   - Displaying the pixel-by-pixel abs val difference as an image.
//   - Computing the mean abs val difference and the rms difference.
//   - Using the difference to fade the image to white or black.
//   - Generating the rank histogram mapping from
//     a pixel difference to the cumulative fraction with an abs pixel.
//     difference equal to or greater than the pixel difference.
//   - Plotting the histogram and saving it as an image.
//   - Extracting any set of values from the histogram.
//   - Tiling an arbitrary set of images into a single image.

#include <cstdio>
#include <cstdlib>
#include <unistd.h>
#include <math.h>
#include "allheaders.h"  /* third_party/leptonica/include/allheaders.h */

static const l_int32    kMaskThreshold = 128;
static const l_int32    kTilesize = 8;
static const l_int32    kBlendtype = L_BLEND_TO_WHITE;

static const char *RefName[] = {
        "no file",
        "/home/dbloomberg/work/imagecomparison/ref_example1.jpg",
        "/home/dbloomberg/work/imagecomparison/ref_example2.jpg",
        "/home/dbloomberg/work/imagecomparison/ref_example3.jpg",
        "/home/dbloomberg/work/imagecomparison/ref_example4.jpg",
        "/home/dbloomberg/work/imagecomparison/ref_example5.jpg"};
static const char *Name[] = {
        "no file",
        "/home/dbloomberg/work/imagecomparison/example1.jpg",
        "/home/dbloomberg/work/imagecomparison/example2.jpg",
        "/home/dbloomberg/work/imagecomparison/example3.jpg",
        "/home/dbloomberg/work/imagecomparison/example4.jpg",
        "/home/dbloomberg/work/imagecomparison/example5.jpg"};
static const char *MaskName[] = {
        "no file",
        "/home/dbloomberg/work/imagecomparison/mask_example1.png",
        "/home/dbloomberg/work/imagecomparison/mask_example2.png",
        "/home/dbloomberg/work/imagecomparison/mask_example3.png",
        "/home/dbloomberg/work/imagecomparison/mask_example4.png",
        "/home/dbloomberg/work/imagecomparison/mask_example5.png"};


main(int argc, char **argv)
{
static char  mainName[] = "comparison";

    if (argc != 2)
        exit(ERROR_INT(" Syntax:  comparison index", mainName, 1));
    l_int32 index = atoi(argv[1]);

        // Store images here, for later tiled output to a single image
    Pixa *pixa = pixaCreate(0);  // def'd in pixabasic.c
    pixDisplayWrite(NULL, -1);

        // Read images and mask
    Pix *pixs1 = pixRead(RefName[index]);  // def'd in readfile.c
    Pix *pixs2 = pixRead(Name[index]);
    Pix *pixm32 = pixRead(MaskName[index]);
    pixDisplayWrite(pixs1, 1);   // def'd in writefile.c
    pixDisplayWrite(pixs2, 1);
    pixaAddPix(pixa, pixs1, L_COPY);   // def'd in pixabasic.c
    pixaAddPix(pixa, pixs2, L_COPY);

        // Convert mask, which happens to be 32 bpp, to 1 bpp
    Pix *pixm = pixConvertTo1(pixm32, kMaskThreshold);  // def'd in pixconv.c

        // Use mask on images, painting 0 (black) through the mask
    pixPaintThroughMask(pixs1, pixm, 0, 0, 0x0);  // def'd in pix3.c
    pixDisplayWrite(pixs1, 1);
    pixPaintThroughMask(pixs2, pixm, 0, 0, 0x0);
    pixDisplayWrite(pixs2, 1);
    pixaAddPix(pixa, pixs1, L_COPY);
    pixaAddPix(pixa, pixs2, L_COPY);

        // Generate difference image and simple difference statistics
    float diff, rmsdiff;
    Pix *pixdiff;
    pixCompareRGB(pixs1, pixs2, L_COMPARE_ABS_DIFF, 0, NULL, &diff, &rmsdiff,
                  &pixdiff);  // def'd in compare.c
    fprintf(stderr, "average abs diff = %7.3f; rmsdiff = %7.3f\n",
            diff, rmsdiff);
    pixDisplayWrite(pixdiff, 1);
    pixaAddPix(pixa, pixdiff, L_COPY);

        // Display changes using fade to white over tiles
    Pix *pixg = pixConvertRGBToLuminance(pixdiff);  // def'd in pixconv.c
    Pix *pixt1 = pixGetAverageTiled(pixg, kTilesize, kTilesize, L_MEAN_ABSVAL);
                                                 // def'd in pix3.c
    Pix *pixt2 = pixExpandReplicate(pixt1, kTilesize);  // def'd in scale.c
    Pix *pixt3 = pixFadeWithGray(pixs1, pixt2, 2.0, kBlendtype);  // in blend.c
    pixDisplayWrite(pixt3, 1);
    pixaAddPix(pixa, pixt3, L_COPY);

        // Generate integrated (rank) histogram that gives the fraction
        // of pixels with a difference GTE a given value, vs that value.
        // You can use this to estimate the parameters you might use
        // to discriminate between 'same' and 'different'.  For example,
        // we show how to extract the fraction in the tail for
        // 10 differences: 5, 10, 15, 20, 25, ... 50
    Numa *na = pixGetGrayHistogramMasked(pixt1, NULL, 0, 0, 1);
    Numa *nax, *nay;
    numaMakeRankFromHistogram(0.0, 1.0, na, 256, &nax, &nay);  // in numafunc.c
    Numa *nac = numaMakeConstant(1.0, 256);  // constant numa, in numafunc.c
    Numa *nar = numaArithOp(NULL, nac, nay, L_ARITH_SUBTRACT);  // in numafunc.c
    GPlot *gplot = gplotCreate("/tmp/junkroot", GPLOT_PNG, 
                               "rank extraction from histo",
                               "pix val", "rank val");  // def'd in gplot.c
    gplotAddPlot(gplot, nax, nar, GPLOT_LINES, "plot 1");  // in gplot.c
    gplotMakeOutput(gplot);   // in gplot.c
    gplotDestroy(&gplot);
    sleep(1);  // gives time to write plot to file
    Pix *pixt4 = pixRead("/tmp/junkroot.png");
    pixDisplayWrite(pixt4, 1);
    pixaAddPix(pixa, pixt4, L_COPY);
    float fval;
    for (l_int32 i = 5; i <= 50; i+=5) {
        numaGetFValue(nar, i, &fval);
        fprintf(stderr, "diff = %d, fract = %7.5f\n", i, fval);
    }

        // Start up gthumb to display the output
  l_int32 ignore = system("/usr/bin/gthumb /tmp/junk_write_display* &");

        // Write the images out, tiled into a single image
    l_int32 outdepth = 32;
    l_int32 tilewidth = 500;
    l_int32 ncols = 3;
    l_int32 background = 0;  // white
    l_int32 spacing = 25;  // between the images
    l_int32 border = 0;  // black border around each image
    Pix *pixt5 = pixaDisplayTiledAndScaled(pixa, outdepth, tilewidth, ncols,
                                           background, spacing, border);
                                              // in pixafunc.c
    pixWrite("/tmp/junktiled", pixt5, IFF_PNG);

        // Clean up so we can use valgrind to verify no errors/leaks
    numaDestroy(&na);
    numaDestroy(&nax);
    numaDestroy(&nay);
    numaDestroy(&nac);
    numaDestroy(&nar);
    pixDestroy(&pixs1);
    pixDestroy(&pixs2);
    pixDestroy(&pixm);
    pixDestroy(&pixm32);
    pixDestroy(&pixdiff);
    pixDestroy(&pixg);
    pixDestroy(&pixt1);
    pixDestroy(&pixt2);
    pixDestroy(&pixt3);
    pixDestroy(&pixt4);
    pixDestroy(&pixt5);
    pixaDestroy(&pixa);
    return 0;
}

